//--------------------------------------------------------------------------
// Animated flame for Adafruit Pro Trinket.  Uses the following parts:
//   - Pro Trinket microcontroller (adafruit.com/product/2010 or 2000)
//     (#2010 = 3V/12MHz for longest battery life, but 5V/16MHz works OK)
//   - Charlieplex LED Matrix Driver (2946)
//   - Charlieplex LED Matrix (2947, 2948, 2972, 2973 or 2974)
//   - 350 mAh LiPoly battery (2750)
//   - LiPoly backpack (2124)
//   - SPDT Slide Switch (805)
//
// This is NOT good "learn from" code for the IS31FL3731; it is "squeeze
// every last byte from the Pro Trinket" code.  If you're starting out,
// download the Adafruit_IS31FL3731 and Adafruit_GFX libraries, which
// provide functions for drawing pixels, lines, etc.  This sketch also
// uses some ATmega-specific tricks and will not run as-is on other chips.
//--------------------------------------------------------------------------

#include <Wire.h>           // For I2C communication
#include <Adafruit_GFX.h>
#include "Adafruit_IS31FL3731_ZKA.h"
#include "candle_data.h"

#define WIDTH      (9)
#define HEIGHT     (16)

Adafruit_IS31FL3731 ledmatrix = Adafruit_IS31FL3731();

uint8_t page = 0;    // Front/back buffer control
uint8_t img[WIDTH * HEIGHT]; // Buffer for rendering image

// CANDLE
const uint8_t *ptr  = CANDLE_ANIM; // Current pointer into animation data

// SWIRL
// The lookup table to make the brightness changes be more visible
uint8_t sweep[] = {1, 2, 3, 4, 6, 8, 10, 15, 20, 30, 40, 60, 60, 40, 30, 20, 15, 10, 8, 6, 4, 3, 2, 1};
uint8_t incr = 0;

const uint8_t BUTTON_PIN = 12;
int8_t anim = 0;
bool button_down = false;
uint16_t bright = 0x0;

uint8_t frame_count = 0;
const uint8_t LED_PIN = 13;

elapsedMillis since_draw;
uint16_t draw_interval_ms = 1000 / 60;

// Begin I2C transmission and write register address (data then follows)
void writeRegister(uint8_t n) {
  Wire.beginTransmission(ISSI_ADDR_DEFAULT);
  Wire.write(n);
  // Transmission is left open for additional writes
}

// Select one of eight IS31FL3731 pages, or Function Registers
void pageSelect(uint8_t n) {
  writeRegister(ISSI_COMMANDREGISTER); // Command Register
  Wire.write(n);       // Page number (or 0xB = Function Registers)
  Wire.endTransmission();
}

inline uint8_t img_idx(uint8_t x, uint8_t y) {
	return y * WIDTH + x;
}

inline void clear_led_matrix() {
	ledmatrix.clear();
	memset(img, 0, sizeof(img) / sizeof(img[0]));
}

void set_anim(uint8_t value) {
	anim = value;
	draw_interval_ms = 1000 / 60;
}

inline uint8_t bright_out(uint8_t value) {
	if (bright < 256) return (value * bright) >> 8;
	return value;
}

void setup() {
	//Serial.begin(9600);

	if (!ledmatrix.begin()) {
		//Serial.println("IS31 not found");
		while (1);
	}
	//Serial.println("IS31 found!");

	pinMode(BUTTON_PIN, INPUT);
	pinMode(LED_PIN, OUTPUT);

	clear_led_matrix();
}

void loop() {

	// 60 FPS or so
	if (since_draw < draw_interval_ms) return;
	since_draw = 0;

	// Display frame rendered on prior pass.  This is done at function start
  // (rather than after rendering) to ensire more uniform animation timing.
  pageSelect(ISSI_BANK_FUNCTIONREG);    // Function registers
  writeRegister(ISSI_REG_PICTUREFRAME); // Picture Display reg
  Wire.write(page);    // Page #
  Wire.endTransmission();

  page ^= 1; // Flip front/back buffer index

	uint8_t new_button_state = digitalRead(BUTTON_PIN);
	if (!button_down && (new_button_state == LOW)) {
		clear_led_matrix();
		bright = 0;
		set_anim((anim + 1) % 2);
	}
	button_down = (new_button_state == LOW);

	if (anim == 0) {
	//if (new_button_state == LOW) {
		// CANDLE
		uint8_t  a, x1, y1, x2, y2, x, y;

		// Then render NEXT frame.  Start by getting bounding rect for new frame:
		a = pgm_read_byte(ptr++);     // New frame X1/Y1
		if(a >= 0x90) {               // EOD marker? (valid X1 never exceeds 8)
			ptr = CANDLE_ANIM;          // Reset animation data pointer to start
			a   = pgm_read_byte(ptr++); // and take first value
		}
		x1 = a >> 4;                  // X1 = high 4 bits
		y1 = a & 0x0F;                // Y1 = low 4 bits
		a  = pgm_read_byte(ptr++);    // New frame X2/Y2
		x2 = a >> 4;                  // X2 = high 4 bits
		y2 = a & 0x0F;                // Y2 = low 4 bits

		for(x=x1; x<=x2; x++) { // Column-major
			for(y=y1; y<=y2; y++) {
				//ledmatrix.drawPixel(y, x, bright_out(pgm_read_byte(ptr++)));
				uint8_t idx = img_idx(x,y);
				img[idx] = bright_out(pgm_read_byte(ptr++));
			}
		}

	} else if (anim == 1) {
		// SWIRL
		// animate over all the pixels, and set the brightness from the sweep table
		incr = (incr + 1) % 24;

		for (uint8_t x = 0; x < WIDTH; x++) {
			for (uint8_t y = 0; y < HEIGHT; y++) {
				//ledmatrix.drawPixel(x, y, bright_out(sweep[(x+y+incr)%24]) );
				uint8_t idx = img_idx(x,y);
				img[idx] = bright_out(sweep[(x+y+incr)%24]);
			}
		}
	}

	// Write img[] to the matrix
	pageSelect(page);    // Select background buffer
	writeRegister(0x24); // First byte of PWM data
	uint8_t i = 0, byteCounter = 0;	// ZKA: byteCounter counts from 0
	for(uint8_t x=0; x<WIDTH; x++) {
		for(uint8_t y=0; y<HEIGHT; y++) {
			Wire.write(img[i++]);      // Write each byte to matrix
			if(++byteCounter >= 32) {  // Every 32 bytes...
				byteCounter = 0;         // ZKA: Added this line, to ensure "Every 32 bytes..."
				Wire.endTransmission();  // end transmission and
				writeRegister(0x24 + i); // start a new one (Wire lib limits)
			}
		}
	}
	Wire.endTransmission();

	// Brightness: Get brighter, towards max
	bright = min(bright+1, 256);

	frame_count = (frame_count + 1) % 60;
	if (frame_count == 0) {
		digitalWrite(LED_PIN, HIGH);
	} else {
		digitalWrite(LED_PIN, LOW);
	}
}
